@@ -1,6 +1,6 @@
 /* -*- mode:C; c-file-style: "bsd" -*- */
 /*
- * Copyright (c) 2008-2014 Yubico AB
+ * Copyright (c) 2008-2015 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,7 +42,16 @@
 #endif
 
 #ifdef YK_DEBUG
-static void _yk_hexdump(void *, int);
+#define _yk_hexdump(buffer, size) \
+	do { \
+		unsigned char *p = buffer; \
+		int i; \
+		fprintf(stderr, "%25s: ", __func__); \
+		for(i = 0; i < size; i++) { \
+			fprintf(stderr, "%02x ", *p++); \
+		} \
+		fprintf(stderr, "\n"); \
+	} while(0)
 #endif
 
 /*
@@ -67,9 +76,17 @@ int yk_release(void)
 
 YK_KEY *yk_open_first_key(void)
 {
-	int pids[] = {YUBIKEY_PID, NEO_OTP_PID, NEO_OTP_CCID_PID};
+	return yk_open_key(0);
+}
 
-	YK_KEY *yk = _ykusb_open_device(YUBICO_VID, pids, sizeof(pids));
+YK_KEY *yk_open_key(int index)
+{
+	int pids[] = {YUBIKEY_PID, NEO_OTP_PID, NEO_OTP_CCID_PID,
+		NEO_OTP_U2F_PID, NEO_OTP_U2F_CCID_PID, YK4_OTP_PID,
+		YK4_OTP_U2F_PID, YK4_OTP_CCID_PID, YK4_OTP_U2F_CCID_PID,
+		PLUS_U2F_OTP_PID};
+
+	YK_KEY *yk = _ykusb_open_device(YUBICO_VID, pids, sizeof(pids) / sizeof(int), index);
 	int rc = yk_errno;
 
 	if (yk) {
@@ -103,27 +120,6 @@ int yk_check_firmware_version(YK_KEY *k)
 
 int yk_check_firmware_version2(YK_STATUS *st)
 {
-	if (!((st->versionMajor == 0 &&
-	       (st->versionMinor == 9 ||
-		st->versionBuild == 9)) ||
-	      (st->versionMajor == 1 &&
-	       (st->versionMinor == 0 ||
-		st->versionMinor == 1 ||
-		st->versionMinor == 2 ||
-		st->versionMinor == 3)) ||
-	      (st->versionMajor == 2 &&
-	       (st->versionMinor == 0 ||
-		st->versionMinor == 1 ||
-		st->versionMinor == 2 ||
-		st->versionMinor == 3 ||
-		st->versionMinor == 4)) ||
-	      (st->versionMajor == 3 &&
-	       (st->versionMinor == 0 ||
-		st->versionMinor == 1 ||
-		st->versionMinor == 2)))) {
-		yk_errno = YK_EFIRMWARE;
-		return 0;
-	}
 	return 1;
 }
 
@@ -187,6 +183,27 @@ int yk_get_serial(YK_KEY *yk, uint8_t slot, unsigned int flags, unsigned int *se
 	return 1;
 }
 
+int yk_get_capabilities(YK_KEY *yk, uint8_t slot, unsigned int flags,
+		unsigned char *capabilities, unsigned int *len)
+{
+	unsigned int response_len = 0;
+
+	if (!yk_write_to_key(yk, SLOT_YK4_CAPABILITIES, capabilities, 0))
+		return 0;
+
+	if (! yk_read_response_from_key(yk, slot, flags,
+					capabilities, *len, 0, /* we have no idea how much data we'll get */
+					&response_len))
+		return 0;
+
+	/* the first data of the capabilities string is the length */
+	response_len = capabilities[0];
+	response_len++;
+
+	*len = response_len;
+	return 1;
+}
+
 static int _yk_write(YK_KEY *yk, uint8_t yk_cmd, unsigned char *buf, size_t len)
 {
 	YK_STATUS stat;
@@ -208,7 +225,8 @@ static int _yk_write(YK_KEY *yk, uint8_t yk_cmd, unsigned char *buf, size_t len)
 	 * want to get the bytes in the status message, but when writing configuration
 	 * we don't expect any data back.
 	 */
-	yk_wait_for_key_status(yk, yk_cmd, 0, WAIT_FOR_WRITE_FLAG, false, SLOT_WRITE_FLAG, NULL);
+	if(!yk_wait_for_key_status(yk, yk_cmd, 0, WAIT_FOR_WRITE_FLAG, false, SLOT_WRITE_FLAG, NULL))
+		return 0;
 
 	/* Verify update */
 
@@ -484,6 +502,9 @@ int yk_wait_for_key_status(YK_KEY *yk, uint8_t slot, unsigned int flags,
 		memset(data, 0, sizeof(data));
 		if (!_ykusb_read(yk, REPORT_TYPE_FEATURE, slot, (char *) &data, FEATURE_RPT_SIZE))
 			return 0;
+#ifdef YK_DEBUG
+		_yk_hexdump(data, FEATURE_RPT_SIZE);
+#endif
 
 		if (last_data != NULL)
 			memcpy(last_data, data, sizeof(data));
@@ -507,7 +528,7 @@ int yk_wait_for_key_status(YK_KEY *yk, uint8_t slot, unsigned int flags,
 				if (! blocking) {
 					/* Extend timeout first time we see RESP_TIMEOUT_WAIT_FLAG. */
 					blocking = 1;
-					max_time_ms += 15000;
+					max_time_ms += 256 * 1000;
 				}
 			} else {
 				/* Reset read mode of Yubikey before aborting. */
@@ -648,12 +669,6 @@ int yk_write_to_key(YK_KEY *yk, uint8_t slot, const void *buf, int bufcount)
 	ptr = (unsigned char *) &frame;
 	end = (unsigned char *) &frame + sizeof(frame);
 
-	/* Initial check that the YubiKey is in a state where it will accept
-	 * a write.
-	 *
-	if (! yk_wait_for_key_status(yk, slot, 0, 1000, false, SLOT_WRITE_FLAG, NULL))
-		return 0;
-	*/
 #ifdef YK_DEBUG
 	fprintf(stderr, "YK_DEBUG: Write %i bytes to YubiKey :\n", bufcount);
 #endif
@@ -700,6 +715,10 @@ int yk_force_key_update(YK_KEY *yk)
 	return 1;
 }
 
+int yk_get_key_vid_pid(YK_KEY *yk, int *vid, int *pid) {
+	return _ykusb_get_vid_pid(yk, vid, pid);
+}
+
 uint16_t yk_endian_swap_16(uint16_t x)
 {
 	static int testflag = -1;
@@ -718,18 +737,3 @@ uint16_t yk_endian_swap_16(uint16_t x)
 
 	return x;
 }
-
-#ifdef YK_DEBUG
-/* Private little hexdump function for debugging */
-static void _yk_hexdump(void *buffer, int size)
-{
-       unsigned char *p = buffer;
-       int i;
-       for (i = 0; i < size; i++) {
-               fprintf(stderr, "%02x ", *p);
-               p++;
-      }
-      fprintf(stderr, "\n");
-      fflush(stderr);
-}
-#endif
