@@ -40,7 +40,6 @@
 #include <ykpers.h>
 #include <yubikey.h> /* To get yubikey_modhex_encode and yubikey_hex_encode */
 #include <ykdef.h>
-#include <ykpers-version.h>
 #include "ykpers-args.h"
 
 #define YUBICO_OATH_VENDOR_ID_HEX	0xe1	/* UB as hex */
@@ -48,6 +47,7 @@
 
 const char *usage =
 "Usage: ykpersonalize [options]\n"
+"-Nkey     use nth key found\n"
 "-u        update configuration without overwriting.  This is only available\n"
 "          in YubiKey 2.3 and later.  EXTFLAG_ALLOW_UPDATE will be set by\n"
 "          default\n"
@@ -64,27 +64,31 @@ const char *usage =
 "-z        delete the configuration in slot 1 or 2.\n"
 "-sFILE    save configuration to FILE instead of key.\n"
 "          (if FILE is -, send to stdout)\n"
-"-iFILE    read configuration from FILE.\n"
+"-iFILE    read configuration from FILE. (only valid for -fycfg)\n"
 "          (if FILE is -, read from stdin)\n"
 "-fformat  set the data format for -s and -i valid values are ycfg or legacy.\n"
 "-a[XXX..] The AES secret key as a 32 (or 40 for OATH-HOTP/HMAC CHAL-RESP)\n"
 "          char hex value (not modhex) (none to prompt for key on stdin)\n"
 "          If -a is not used a random key will be generated.\n"
-"-cXXX..   A 12 char hex value (not modhex) to use as access code for programming\n"
+"-c[XXX..] A 12 char hex value (not modhex) to use as access code for programming\n"
 "          (this does NOT SET the access code, that's done with -oaccess=)\n"
+"          If -c is provided without argument a code is prompted for\n"
 "-nXXX..   Write NDEF URI to YubiKey NEO, must be used with -1 or -2\n"
 "-tXXX..   Write NDEF text to YubiKey NEO, must be used with -1 or -2\n"
-"-mMODE    Set the USB device configuration of the YubiKey NEO.\n"
-"          See the manpage for details\n"
-"-S0605..  Set the scanmap to use with the YubiKey NEO. Must be 45 unique bytes,\n"
-"          in hex.  Use with no argument to reset to the default.\n"
+"-mMODE    Set the USB device configuration of the YubiKey.\n"
+"          See the manpage for details. This is for YubiKey 3.0 and newer only.\n"
+"-S0605..  Set the scanmap to use with the YubiKey. Must be 45 unique bytes,\n"
+"          in hex.  Use with no argument to reset to the default. This is for\n"
+"          YubiKey 3.0 and newer only.\n"
 "-oOPTION  change configuration option.  Possible OPTION arguments are:\n"
 "          fixed=xxxxxxxxxxx   The public identity of key, in MODHEX.\n"
-"                              This is 0-16 characters long.\n"
-"          uid=xxxxxx          The uid part of the generated ticket, in HEX.\n"
+"                              This is 0-32 characters long.\n"
+"          uid[=xxxxxx]        The uid part of the generated ticket, in HEX.\n"
 "                              MUST be 12 characters long.\n"
-"          access=xxxxxxxxxxx  New access code to set, in HEX.\n"
+"                              If argument is omitted uid is prompted for on stdin.\n"
+"          access[=xxxxxx]     New access code to set, in HEX.\n"
 "                              MUST be 12 characters long.\n"
+"                              If argument is omitted code is prompted for on stdin.\n"
 "          oath-imf=IMF        OATH Initial Moving Factor to use.\n"
 "          oath-id[=h:OOTT...] OATH Token Identifier (none for serial-based)\n"
 "\n"
@@ -155,14 +159,13 @@ const char *usage =
 "-V        tool version\n"
 "-h        help (this text)\n"
 ;
-const char *optstring = "u12xza::c:n:t:hi:o:s:f:dvym:S::V";
+const char *optstring = ":u12xza:c:n:t:hi:o:s:f:dvym:S:VN:";
 
 static int _set_fixed(char *opt, YKP_CONFIG *cfg);
 static int _format_decimal_as_hex(uint8_t *dst, size_t dst_len, uint8_t *src);
 static int _format_oath_id(uint8_t *dst, size_t dst_len, uint8_t vendor, uint8_t type, uint32_t mui);
-static int _set_oath_id(char *opt, YKP_CONFIG *cfg, YK_KEY *yk, YK_STATUS *st);
 
-static int hex_modhex_decode(unsigned char *result, size_t *resultlen,
+int hex_modhex_decode(unsigned char *result, size_t *resultlen,
 			     const char *str, size_t strl,
 			     size_t minsize, size_t maxsize,
 			     bool primarily_modhex)
@@ -217,6 +220,24 @@ void report_yk_error(void)
 	}
 }
 
+static int prompt_for_data(const char *prompt, char **data) {
+	fprintf(stderr, "%s", prompt);
+	fflush(stderr);
+	*data = calloc(257, sizeof(char));
+	if(!fgets(*data, 256, stdin)) {
+			fprintf(stderr, "Error reading from stdin\n");
+			perror ("fgets");
+			return 1;
+	}
+	if((*data)[strlen(*data) - 1] == '\n') {
+			(*data)[strlen(*data) - 1] = '\0';
+	}
+	return 0;
+}
+
+extern char *optarg;
+extern int optind;
+
 /*
  * Parse all arguments supplied to this program and turn it into mainly
  * a YKP_CONFIG (but return some other parameters as well, like
@@ -224,20 +245,20 @@ void report_yk_error(void)
  *
  * Done in this way to be testable (see tests/test_args_to_config.c).
  */
-int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
+int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, char *oathid,
 		   const char **infname, const char **outfname,
 		   int *data_format, bool *autocommit,
 		   YK_STATUS *st, bool *verbose, bool *dry_run,
-		   unsigned char *access_code, bool *use_access_code,
-		   char *keylocation, char *ndef_type, char *ndef,
+		   char **access_code, char **new_access_code,
+		   char *ndef_type, char *ndef,
 		   unsigned char *usb_mode, bool *zap,
 		   unsigned char *scan_bin, unsigned char *cr_timeout,
-		   unsigned char *autoeject_timeout, int *num_modes_seen,
+		   unsigned short *autoeject_timeout, int *num_modes_seen,
 		   int *exit_code)
 {
 	int c;
+	char keylocation = 0;
 	const char *aeshash = NULL;
-	bool new_access_code = false;
 	bool slot_chosen = false;
 	bool mode_chosen = false;
 	bool option_seen = false;
@@ -378,32 +399,25 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 			}
 			break;
 		case 'a':
-			if(optarg) {
-				aeshash = optarg;
-				*keylocation = 1;
+			if(optarg[0] == '-') {
+				keylocation = 2;
+				optind--;
 			} else {
-				*keylocation = 2;
+				aeshash = optarg;
+				keylocation = 1;
 			}
 			break;
-		case 'c': {
-			size_t access_code_len = 0;
-			int rc = hex_modhex_decode(access_code, &access_code_len,
-						   optarg, strlen(optarg),
-						   12, 12, false);
-			if (rc <= 0) {
-				fprintf(stderr,
-					"Invalid access code string: %s\n",
-					optarg);
-				*exit_code = 1;
-				return 0;
+		case 'c':
+			if(optarg[0] == '-') {
+				optind--;
+				if(prompt_for_data(" Access code, 6 bytes (12 characters hex) : ", access_code) != 0) {
+					*exit_code = 1;
+					return 0;
+				}
+			} else {
+				*access_code = strdup(optarg);
 			}
-			if (!new_access_code)
-				ykp_set_access_code(cfg,
-						    access_code,
-						    access_code_len);
-			*use_access_code = true;
 			break;
-		}
 		case 't':
 			*ndef_type = 'T';
 		case 'n': {
@@ -436,8 +450,9 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 				*exit_code = 1;
 				return 0;
 			}
-			unsigned char mode, crtime, autotime;
-			int matched = sscanf(optarg, "%hhx:%hhd:%hhd", &mode, &crtime, &autotime);
+			unsigned char mode, crtime;
+			unsigned short autotime;
+			int matched = sscanf(optarg, "%hhx:%hhd:%hd", &mode, &crtime, &autotime);
 			if(matched > 0) {
 				*usb_mode = mode;
 				if(matched > 1) {
@@ -465,7 +480,7 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 					*exit_code = 1;
 					return 0;
 				}
-				if(optarg) {
+				{
 					size_t scanbinlen;
 					size_t scanlen = strlen (optarg);
 					int rc = hex_modhex_decode(scan_bin, &scanbinlen,
@@ -480,8 +495,6 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 						*exit_code = 1;
 						return 0;
 					}
-				} else {
-					memset(scan_bin, 0, scanlength);
 				}
 				scan_map_seen = true;
 			}
@@ -503,48 +516,53 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 					return 0;
 				}
 			}
-			else if (strncmp(optarg, "uid=", 4) == 0) {
-				const char *uid = optarg+4;
-				size_t uidlen = strlen (uid);
+			else if (strncmp(optarg, "uid", 3) == 0) {
+				char *uid = optarg+4;
+				size_t uidlen;
 				unsigned char uidbin[256];
 				size_t uidbinlen = 0;
-				int rc = hex_modhex_decode(uidbin, &uidbinlen,
-							   uid, uidlen,
-							   12, 12, false);
+				int rc;
+				char *uidtmp = NULL;
+
+				if(strncmp(optarg, "uid=", 4) != 0) {
+					if(prompt_for_data(" Private ID, 6 bytes (12 characters hex) : ", &uidtmp) != 0) {
+						*exit_code = 1;
+						return 0;
+					}
+					uid = uidtmp;
+				}
+
+				uidlen = strlen(uid);
+				rc = hex_modhex_decode(uidbin, &uidbinlen,
+						uid, uidlen,
+						12, 12, false);
 				if (rc <= 0) {
 					fprintf(stderr,
-						"Invalid uid string: %s\n",
-						uid);
+							"Invalid uid string: %s\n",
+							uid);
 					*exit_code = 1;
 					return 0;
 				}
+
+				free(uidtmp);
 				/* for OATH-HOTP and CHAL-RESP, uid is not applicable */
 				if (ykp_get_tktflag_OATH_HOTP(cfg) || ykp_get_tktflag_CHAL_RESP(cfg)) {
 					fprintf(stderr,
-						"Option uid= not valid with -ooath-hotp or -ochal-resp.\n"
-						);
+							"Option uid= not valid with -ooath-hotp or -ochal-resp.\n"
+							);
 					*exit_code = 1;
 					return 0;
 				}
 				ykp_set_uid(cfg, uidbin, uidbinlen);
 			}
 			else if (strncmp(optarg, "access=", 7) == 0) {
-				const char *acc = optarg+7;
-				size_t acclen = strlen (acc);
-				unsigned char accbin[256];
-				size_t accbinlen = 0;
-				int rc = hex_modhex_decode (accbin, &accbinlen,
-							    acc, acclen,
-							    12, 12, false);
-				if (rc <= 0) {
-					fprintf(stderr,
-						"Invalid access code string: %s\n",
-						acc);
+				*new_access_code = strdup(optarg + 7);
+			}
+			else if (strncmp(optarg, "access", 6) == 0) {
+				if(prompt_for_data(" New access code, 6 bytes (12 characters hex) : ", new_access_code) != 0) {
 					*exit_code = 1;
 					return 0;
 				}
-				ykp_set_access_code(cfg, accbin, accbinlen);
-				new_access_code = true;
 			}
 #define TKTFLAG(o, f)							\
 			else if (strcmp(optarg, o) == 0) {		\
@@ -626,10 +644,7 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 				}
 			}
 			else if (strncmp(optarg, "oath-id=", 8) == 0 || strcmp(optarg, "oath-id") == 0) {
-				if (_set_oath_id(optarg, cfg, yk, st) != 1) {
-					*exit_code = 1;
-					return 0;
-				}
+				strcpy(oathid, optarg);
 			}
 
 #define EXTFLAG(o, f)							\
@@ -671,9 +686,34 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 			*autocommit = true;
 			break;
 		case 'V':
-			fputs(YKPERS_VERSION_STRING "\n", stderr);
-			*exit_code = 0;
-			return 0;
+		case 'N':
+			continue;
+		case ':':
+			switch(optopt) {
+				case 'S':
+					{
+						size_t scanlength = strlen(SCAN_MAP);
+						if(slot_chosen || swap_seen || update_seen || option_seen || ndef_seen || *zap || usb_mode_seen) {
+							fprintf(stderr, "Scanmap (-S) can not be combined with other options.\n");
+							*exit_code = 1;
+							return 0;
+						}
+						memset(scan_bin, 0, scanlength);
+						scan_map_seen = true;
+						if (!ykp_configure_command(cfg, SLOT_SCAN_MAP))
+							return 0;
+						continue;
+					}
+				case 'a':
+					keylocation = 2;
+					continue;
+				case 'c':
+					if(prompt_for_data(" Access code, 6 bytes (12 characters hex) : ", access_code) != 0) {
+						*exit_code = 1;
+						return 0;
+					}
+					continue;
+			}
 		case 'h':
 		default:
 			fputs(usage, stderr);
@@ -683,7 +723,11 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 	}
 
 	if (!slot_chosen && !ndef_seen && !swap_seen && !usb_mode_seen && !scan_map_seen) {
-		fprintf(stderr, "A slot must be chosen with -1 or -2.\n");
+		if (argc == 1) {
+			fputs(usage, stderr);
+		} else {
+			fprintf(stderr, "A slot must be chosen with -1 or -2.\n");
+		}
 		*exit_code = 1;
 		return 0;
 	}
@@ -707,19 +751,78 @@ int args_to_config(int argc, char **argv, YKP_CONFIG *cfg, YK_KEY *yk,
 		}
 	}
 
-	if (*keylocation == 1) {
-		bool long_key_valid = ykp_get_supported_key_length(cfg) == 20 ? true : false;
+	if (! *zap && (ykp_command(cfg) == SLOT_CONFIG || ykp_command(cfg) == SLOT_CONFIG2)) {
+		size_t key_bytes = (size_t)ykp_get_supported_key_length(cfg);
 		int res = 0;
+		char *key_tmp = NULL;
+		char keybuf[20];
 
-		if (long_key_valid && strlen(aeshash) == 40) {
-			res = ykp_HMAC_key_from_hex(cfg, aeshash);
+		if(keylocation == 2) {
+			const char *prompt = " AES key, 16 bytes (32 characters hex) : ";
+			if (key_bytes == 20) {
+				prompt = " HMAC key, 20 bytes (40 characters hex) : ";
+			}
+			if (prompt_for_data(prompt, &key_tmp) != 0) {
+				*exit_code = 1;
+				return 0;
+			}
+			aeshash = key_tmp;
+			keylocation = 1;
+		}
+
+		if(keylocation == 0) {
+			const char *random_places[] = {
+				"/dev/srandom",
+				"/dev/urandom",
+				"/dev/random",
+				0
+			};
+			const char **random_place;
+			size_t read_bytes = 0;
+
+			for (random_place = random_places; *random_place; random_place++) {
+				FILE *random_file = fopen(*random_place, "r");
+				if (random_file) {
+					read_bytes = 0;
+
+					while (read_bytes < key_bytes) {
+						size_t n = fread(&keybuf[read_bytes], 1,
+								key_bytes - read_bytes, random_file);
+						read_bytes += n;
+					}
+
+					fclose(random_file);
+					break;
+				}
+			}
+			if(read_bytes < key_bytes) {
+				ykp_errno = YKP_ENORANDOM;
+				*exit_code = 1;
+				return 0;
+			}
 		} else {
-			res = ykp_AES_key_from_hex(cfg, aeshash);
+			size_t key_len = 0;
+			int rc = hex_modhex_decode((unsigned char *)keybuf, &key_len, aeshash, strlen(aeshash), key_bytes * 2, key_bytes * 2, false);
+
+			free(key_tmp);
+
+			if(rc <= 0) {
+				fprintf(stderr, "Invalid key string\n");
+				*exit_code = 1;
+				return 0;
+			}
+		}
+
+		if (key_bytes == 20) {
+			res = ykp_HMAC_key_from_raw(cfg, keybuf);
+		} else {
+			res = ykp_AES_key_from_raw(cfg, keybuf);
 		}
 
 		if (res) {
-			fprintf(stderr, "Bad %s key: %s\n", long_key_valid ? "HMAC":"AES", aeshash);
+			fprintf(stderr, "Bad %s key: %s\n", key_bytes == 20 ? "HMAC":"AES", aeshash);
 			fflush(stderr);
+			*exit_code = 1;
 			return 0;
 		}
 	}
@@ -784,7 +887,7 @@ static int _format_oath_id(uint8_t *dst, size_t dst_len, uint8_t vendor, uint8_t
 	return 1;
 }
 
-static int _set_oath_id(char *opt, YKP_CONFIG *cfg, YK_KEY *yk, YK_STATUS *st) {
+int set_oath_id(char *opt, YKP_CONFIG *cfg, YK_KEY *yk, YK_STATUS *st) {
 	/* For details, see YubiKey Manual 2010-09-16 section 5.3.4 - OATH-HOTP Token Identifier */
 	if (!ykp_get_tktflag_OATH_HOTP(cfg)) {
 		fprintf(stderr,
